#  tests for sympy-1.6-py38_0 (this is a generated file);
print('===== testing package: sympy-1.6-py38_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import sympy
from sympy import Rational


a = Rational(14, 4)
assert str(a) == '7/2'
#  --- run_test.py (end) ---

print('===== sympy-1.6-py38_0 OK =====');
print("import: 'sympy'")
import sympy

